#ifndef UNICODE
#define UNICODE
#endif
#ifndef _UNICODE
#define _UNICODE
#endif
#define WIN32_LEAN_AND_MEAN
#define _WIN32_WINNT 0x500
#define WIN32_NO_STATUS
#include <windows.h>
#undef WIN32_NO_STATUS
#include <ntstatus.h>
#include <winioctl.h>
#include <winternl.h>
#include <psapi.h>
#include <locale.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include "ioctlcmd.h"
#include "senable.h"

const TCHAR *myname;

enum {
	ACT_LOAD,
	ACT_INSTALL,
	ACT_DELETE,
};

void PrintWin32Error(LPCTSTR message, DWORD code)
{
	PVOID pv;

	FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS |
	              FORMAT_MESSAGE_FROM_SYSTEM, NULL, code,
	              MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
	              (LPTSTR)&pv, 0, NULL);
	_tprintf(_T("%s (%d): %s"), message, code, pv);
	LocalFree(pv);
}

BOOL InstallDriver(SC_HANDLE hSCM, LPCTSTR name, LPCTSTR exe, BOOL bootStart)
{
	SC_HANDLE hService;

	_tprintf(_T("Install..."));
	hService = CreateService(hSCM, name, name, SERVICE_ALL_ACCESS,
	                         SERVICE_KERNEL_DRIVER,
	                         bootStart ? SERVICE_BOOT_START : SERVICE_DEMAND_START,
	                         SERVICE_ERROR_NORMAL, exe,
	                         _T("Filter"), NULL, NULL, NULL, NULL);
	if (!hService) {
		//PrintWin32Error(_T("CreateService failed"), GetLastError());
		return FALSE;
	}

	CloseServiceHandle(hService);

	return TRUE;
}

BOOL ChangeDriver(SC_HANDLE hSCM, LPCTSTR name, LPCTSTR exe, BOOL bootStart)
{
	SC_HANDLE hService;
	BOOL r;

	_tprintf(_T("Change..."));
	hService = OpenService(hSCM, name, SERVICE_ALL_ACCESS);
	if (!hService)
		return FALSE;

	r = ChangeServiceConfig(hService,
	                        SERVICE_KERNEL_DRIVER,
	                        bootStart ? SERVICE_BOOT_START : SERVICE_DEMAND_START,
	                        SERVICE_ERROR_NORMAL, exe,
	                        _T("Filter"), NULL, NULL, NULL, NULL,
	                        name);

	CloseServiceHandle(hService);

	return r;
}

BOOL StartDriver(SC_HANDLE hSCM, LPCTSTR name)
{
	SC_HANDLE hService;
	BOOL r;

	_tprintf(_T("Start..."));
	hService = OpenService(hSCM, name, SERVICE_ALL_ACCESS);
	if (hService == NULL)
		return FALSE;

	r = StartService(hService, 0, NULL) ||
	    GetLastError() == ERROR_SERVICE_ALREADY_RUNNING ||
	    GetLastError() == ERROR_SERVICE_DISABLED;
if (!r) PrintWin32Error(_T("StartService failed"), GetLastError());


	CloseServiceHandle(hService);

	return r;
}



BOOL OpenDevice(LPCTSTR name, PHANDLE phDevice)
{
	TCHAR completeDeviceName[64];
	HANDLE hDevice;

	if (GetVersion() & 0xFF >= 5) {
		wsprintf(completeDeviceName, TEXT("\\\\.\\Global\\%s"), name);
	} else {
		wsprintf(completeDeviceName, TEXT("\\\\.\\%s"), name);
	}
	hDevice = CreateFile(completeDeviceName, GENERIC_READ | GENERIC_WRITE,
	                     0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hDevice == INVALID_HANDLE_VALUE)
		return FALSE;

	if (phDevice)
		*phDevice = hDevice;
	else
		CloseHandle(hDevice);

	return TRUE;
}

BOOL StopDriver(SC_HANDLE hSCM, LPCTSTR name)
{
	SC_HANDLE hService;
	BOOL r;
	SERVICE_STATUS status;

	_tprintf(_T("Stop..."));
	hService = OpenService(hSCM, name, SERVICE_ALL_ACCESS);
	if (hService == NULL)
		return FALSE;

	r = ControlService(hService, SERVICE_CONTROL_STOP, &status);
if (!r) PrintWin32Error(_T("ControlService failed"), GetLastError());

	CloseServiceHandle(hService);

	return r;
}

BOOL RemoveDriver(SC_HANDLE hSCM, LPCTSTR name)
{
	SC_HANDLE hService;
	BOOL r;

	_tprintf(_T("Delete..."));
	hService = OpenService(hSCM, name, SERVICE_ALL_ACCESS);
	// T[rX݂Ȃꍇ̓G[ɂȂ
	if (hService == NULL) {
		//PrintWin32Error(_T("OpenService failed"), GetLastError());
		return GetLastError() == ERROR_SERVICE_DOES_NOT_EXIST;
	}

	// 폜\ς݂̏ꍇ̓G[ɂȂ
	r = DeleteService(hService) || GetLastError() == ERROR_SERVICE_MARKED_FOR_DELETE;
if (!r) PrintWin32Error(_T("DeleteService failed"), GetLastError());

	CloseServiceHandle(hService);

	return r;
}

BOOL CanUnload(LPCTSTR name, LPBOOL inuse)
{
	HANDLE hDevice;
	ULONG irpcount;
	BOOL b = FALSE;

	if (inuse)
		*inuse = FALSE;

	if (!OpenDevice(name, &hDevice)) {
		_tprintf(_T("foCX %s J܂: %lu\n"), name, GetLastError());
		return FALSE;
	}

	// hCoA[hł邩ǂ
	if (!DeviceIoControl(hDevice, IOCTL_SYMLINK_UNLOADQUERY,
						NULL, 0, NULL, 0,
						&irpcount, NULL ) ) {
		_tprintf(_T("Couldn't access device driver: %lu\n"), GetLastError());
	} else if (irpcount) {
		_tprintf(_T("gp̂SymLinkhCoA[hł܂B\n"));
		if (inuse)
			*inuse = TRUE;
	} else {
		b = TRUE;
	}

	CloseHandle(hDevice);

	return b;
}

BOOL UnloadDeviceDriver(LPCTSTR name, LPBOOL inuse, LPBOOL removed)
{
	SC_HANDLE hSCM;
	BOOL r;

	hSCM = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);

	if (CanUnload(name, inuse))
		StopDriver(hSCM, name);
	// A[h̉ۂɂ炸폜͎݂
	r = RemoveDriver(hSCM, name);
	if (removed)
		*removed = r;

	CloseServiceHandle(hSCM);

	return TRUE;
}

BOOL LoadDeviceDriver(LPCTSTR name, LPCTSTR path, BOOL bootStart, PHANDLE phDevice,
PDWORD error, LPBOOL inuse)
{
	SC_HANDLE hSCM;
	BOOL r;

	hSCM = OpenSCManager(NULL, NULL, SC_MANAGER_ALL_ACCESS);

	if (CanUnload(name, inuse))
		StopDriver(hSCM, name);

	if (!ChangeDriver(hSCM, name, path, bootStart)) {
		// ȑÕCX^X菜
		RemoveDriver(hSCM, name);
		// CXg[͖̐: łɃCXg[Ă邩Ȃ
		InstallDriver(hSCM, name, path, bootStart);
	}

	// Jn͖̐: łɊJnĂ邩Ȃ
	StartDriver(hSCM, name);

	// J邱ƂmF
	r = OpenDevice(name, phDevice);
	*error = GetLastError();
	CloseServiceHandle(hSCM);

	return r;
}

//
// hCõt@CpX猟āA
// w肵fBNgɃRs[B
//
BOOL copydriver(LPCTSTR driverPath)
{
	TCHAR Path[MAX_PATH];
	TCHAR *filePart;
	WIN32_FIND_DATA wfd;
	HANDLE hFind;

	GetCurrentDirectory(sizeof(Path), Path);
	_stprintf(Path + lstrlen(Path), _T("\\%s"), SYS_FILE);

	hFind = FindFirstFile(Path, &wfd);
	if (hFind == INVALID_HANDLE_VALUE) {
		if (!SearchPath(NULL, SYS_FILE, NULL, sizeof(Path), Path, &filePart)) {
			_tprintf(_T("%s not found.\n"), SYS_FILE);
			return FALSE;
		}
	} else FindClose(hFind);

	if (!CopyFile(Path, driverPath, FALSE)) {
		_tprintf(_T("%s  %s ɃRs[ł܂B\n")
			_T("ǗҌĂ邩mFĂB\n"),
			SYS_FILE, driverPath);
		return FALSE;
	}

	return TRUE;
}

BOOL opendriver(LPHANDLE phDriver, DWORD action)
{
	DWORD version, cb;
	DWORD error;
	TCHAR systemPath[MAX_PATH];
	TCHAR driverPath[MAX_PATH];
	BOOL bootStart = action == ACT_INSTALL;
	BOOL inuse;

	if (!GetSystemDirectory(systemPath, sizeof(systemPath))) {
		_tprintf(_T("Could not find system directory"));
		return FALSE;
	}
	_stprintf(driverPath, _T("%s\\drivers\\%s"), systemPath, SYS_FILE);

	if (action == ACT_DELETE) {
		BOOL removed;

		UnloadDeviceDriver(SYS_NAME, &inuse, &removed);
		if (inuse)
			_tprintf(_T("폜ɂWindows̍ċNKvłB\n"));
		SetFileAttributes(driverPath, FILE_ATTRIBUTE_NORMAL);
		// 炩̗RŃT[rX̓o^폜Ɏsꍇ
		// t@C폜Ȃ
		if (removed)
			DeleteFile(driverPath);
		return TRUE;
	}

	// CDlbg[NhCusł悤
	// hCo system32\drivers փRs[
	// u[g[hhCoƂēo^邽߂ɂKv
	if (!copydriver(driverPath))
		return FALSE;
	SetFileAttributes(driverPath, FILE_ATTRIBUTE_NORMAL);

	if (!LoadDeviceDriver(SYS_NAME, driverPath, bootStart, phDriver, &error, &inuse)) {
		if (!LoadDeviceDriver(SYS_NAME, driverPath, bootStart, phDriver, &error, &inuse)) {
			_tprintf(_T("Error loading %s: %d"), SYS_NAME, error);
			DeleteFile(driverPath);
			return FALSE;
		}
	}

	// hCõo[W͐?
	if (!DeviceIoControl(*phDriver, IOCTL_SYMLINK_VERSION,
	                     NULL, 0, &version, sizeof(version), &cb, NULL)) {
		PrintWin32Error(_T("DeviceIoControl failed"), GetLastError());
		CloseHandle(*phDriver);
		return FALSE;
	} else if (version < SYMLINKVERSION) {
		_tprintf(SYS_NAME _T(" ̃o[W܂B\n"));
		if (inuse)
			_tprintf(_T("hCoXVɂWindows̍ċNKvłB\n"));
		return FALSE;
	}

	_tprintf(_T("driver version: %d.%02d\n"), version >> 8, version & 0xff);

	return TRUE;
}

int _tmain(int argc, TCHAR **argv)
{
	HANDLE hDriver = INVALID_HANDLE_VALUE;
	DWORD action;

	_tsetlocale(LC_ALL, _T(""));
	myname = argv[0];

	action = ACT_LOAD;
	if (argc > 1) {
		if (!_tcscmp(argv[1], _T("install")))
			action = ACT_INSTALL;
		if (!_tcscmp(argv[1], _T("delete")))
			action = ACT_DELETE;
	}
	if (!opendriver(&hDriver, action))
		return EXIT_FAILURE;

	CloseHandle(hDriver);
	return 0;
}
